# -*- coding: utf-8 -*-
import numpy as np
import unittest as ut
from model import WorldInput, WorldOutput, Sphere

class WorldFile(object):
    
    world = None
    file_path = None
    file_object = None
    current_state_offset = 0
    
    def __init__(self, world, file_path):
        self.world = world
        self.file_path = file_path
        self.reset_state()
    
    def load_step(self):
        return self.consume_state()
    
    def save_step(self):
        self.write_state()
    
    def save(self):
        if self.file_object is not None:
            self.file_object.flush()

    def load_all_steps(self):
        while self.consume_state():
            pass

    def write_header(self):
        f = self.open_file()
        for wo in self.world.world_objects.itervalues():
            line = '#' + str(wo.id) + ',' + type(wo).__name__
            f.write(line + '\n')
        self.current_state_offset = f.tell()        
        
    def consume_header(self):
        self.current_state_offset = 0
        f = self.open_file()
        
        # Read lines starting with '#'
        while True:
            line = f.readline().strip()
            if line == '':
                break
            # Check for and remove '#'
            if not line[0] == '#':
                break
            line = line[1:]
            # Split line into name and type
            elements = line.split(',', 2)
            if len(elements) != 2:
                raise ValueError('Unexpected number of elements in header.', 
                                 len(elements), 
                                 2, # Expected number
                                 self.current_state_offset, 
                                 line)
            object_name = elements[0]
            object_type = elements[1]
            # Create object
            if object_type == 'Sphere':
                self.world.world_objects[object_name] = FileSphere()
            else:
                raise ValueError('Unknown object type.', 
                                 object_type,  
                                 self.current_state_offset, 
                                 line)
                
            # Save offset of next line
            self.current_state_offset = f.tell()
            
        # Reset world time
        self.world.time = 0
    
    def write_state(self):
        f = self.open_file()
        for key, wo in self.world.world_objects.items():
            line = str(self.world.time) + ',' + key + ',' + wo.save_to_string()
            f.write(line + '\n')
        self.current_state_offset = f.tell()
        
    def consume_state(self):
        f = self.open_file()
        
        # Read line until timestamp changes or EOF
        state_timestamp = -1
        while True:
            # Remove \n
            line = f.readline().strip()
            if line == '':
                break
            # Split line into time, name, and object state
            elements = line.split(',', 2)
            if len(elements) != 3:
                raise ValueError('Unexpected number of elements in state.', 
                                 len(elements), 
                                 3, # Expected number
                                 self.current_state_offset, 
                                 line)
            timestamp = float(elements[0])
            object_name = elements[1]
            object_state = elements[2]
            # Set initital timestamp or check if it still matches
            if state_timestamp < 0:
                state_timestamp = timestamp
            elif state_timestamp != timestamp:
                # Timestamp changed -> state has been fully consumed
                break
            # Check object_name
            if not object_name in self.world.world_objects.keys():
                raise ValueError('Unknown world object.', 
                                 object_name,
                                 self.current_state_offset, 
                                 line)
            # Load object state
            wo = self.world.world_objects[object_name]
            wo.load_from_string(object_state)
            
            # Save offset of next line
            self.current_state_offset = f.tell()
        
        # Set world time to state timestamp
        if state_timestamp >= 0:
            self.world.time = state_timestamp
            return True
        
        return False
        
    def reset_state(self):
        self.consume_header()
        self.consume_state()
        
    def open_file(self):
        # Reopen file if closed
        if self.file_object is None or self.file_object.closed:
            self.file_object = open(self.file_path, 'a+')
        # Seek to current state offset
        self.file_object.seek(self.current_state_offset)
        return self.file_object
    
    def close_file(self):
        if self.file_object is not None:
            self.file_object.close()
    
class WorldFileInput(WorldInput):
    
    world_file = None
    
    def __init__(self, world, file_path):
        self.world_file = WorldFile(world, file_path)
        
    def start(self):
        pass
    
    def step(self):
        self.world_file.load_step()
        
    def stop(self):
        self.world_file.close_file()
        
class WorldFileOutput(WorldOutput):
    
    world_file = None
    
    def __init__(self, world, file_path):
        self.world_file = WorldFile(world, file_path)
        self.world_file.load_all_steps()
        
    def start(self):
        pass
    
    def step(self):
        self.world_file.save_step()
        
    def stop(self):
        self.world_file.close_file()
    
class FileSphere(Sphere):
    
    def __init__(self,
                 mass = 0.0,
                 position = np.array([0.0, 0.0, 0.0]), 
                 velocity = np.array([0.0, 0.0, 0.0]),
                 force = np.array([0.0, 0.0, 0.0]),
                 radius = 0.0):
        super(FileSphere, self).__init__(mass, position, velocity, force, radius)
    
    def save_to_string(self):
        return ','.join(map(str,[
                self.position[0],
                self.position[1],
                self.position[2],
                self.velocity[0],
                self.velocity[1],
                self.velocity[2],
                self.force[0],
                self.force[1],
                self.force[2],
                self.mass,
                self.radius]))
   
    def load_from_string(self, state):
        elements = state.split(',')
        self.position = np.array([float(elements[0]), float(elements[1]), float(elements[2])])
        self.velocity = np.array([float(elements[3]), float(elements[4]), float(elements[5])])
        self.force = np.array([float(elements[6]), float(elements[7]), float(elements[8])])
        self.mass = float(elements[9])
        self.radius = float(elements[10])
        
    def __eq__(self, other):
        if type(other) is type(self):
            return (self.position == other.position).all() \
                and (self.velocity == other.velocity).all() \
                and (self.force == other.force).all() \
                and (self.mass == other.mass) \
                and (self.radius == other.radius)
        return False
    
    def __ne__(self, other):
        return not self.__eq__(other)

#Tests    

class TestLoadSave(ut.TestCase):

    def test_save_load_world(self):
        
        path = 'test/test_save_load_world.csv'
        try:
            world = FileWorld(path)
            self.assertEqual(len(world.world_objects), 3)
            
            earth = FileSphere(name='earth', mass=10, position=np.array([1,2,3]), velocity=np.array([4,5,6]), force=np.array([7,8,9]), radius=11)
            self.assertEqual(world.world_objects['earth'], earth)
        finally:
            pass
            #if os.path.exists(path):
            #    os.remove(path)
    
    def test_save_load_sphere(self):
        
        m1 = 34.12
        p1 = np.array([12.34, -56.78, -9.123])
        v1 = np.array([-4.5, -6.789, 1.2])
        f1 = np.array([34.567, 8.9, -12.34])
        r1 = 56.789
    
        sp1 = FileSphere(name='test', mass=m1, position=p1, velocity=v1, force=f1, radius=r1)
        
        state1 = sp1.save_to_string()
    
        sp2 = FileSphere(name='test')
        sp2.load_from_string(state1)
        
        self.assertEqual(sp1, sp2)
                
    def test_save_sphere(self):
        
        m1 = -98.23
        p1 = np.array([24.78, -3432.7945848, 0.0])
        v1 = np.array([-834.1, 546.34, 9.23])
        f1 = np.array([8457.8, -0.00001, -43434.9399822])
        r1 = 92937.0
    
        sp1 = FileSphere(name='test', mass=m1, position=p1, velocity=v1, force=f1, radius=r1)
        
        state1 = sp1.save_to_string()
        
        self.assertEqual(state1, "24.78,-3432.7945848,0.0,-834.1,546.34,9.23,8457.8,-1e-05,-43434.9399822,-98.23,92937.0")
            
    def test_load_sphere(self):
        
        m1 = 23876.0
        p1 = np.array([-23.344, -23.0000023, 0.2])
        v1 = np.array([-7.1, 0.00123, -9.999])
        f1 = np.array([8.2, 1.0023001, -99999999999.9])
        r1 = 0.02323
    
        state1 = "-23.344,-23.0000023,0.2,-7.1,0.00123,-9.999,8.2,1.0023001,-99999999999.9,23876.0,0.02323"
        
        sp2 = FileSphere(name='test')
        sp2.load_from_string(state1)
    
        self.assertTrue((sp2.position == p1).all())
        self.assertTrue((sp2.velocity == v1).all())
        self.assertTrue((sp2.force == f1).all())
        self.assertEqual(sp2.mass, m1)
        self.assertEqual(sp2.radius, r1)
        
if __name__ == '__main__':
    ut.main()