# -*- coding: utf-8 -*-
from model import World, WorldOutput
from file_model import WorldFileInput, WorldFileOutput
from physics_euler import PhysicsEngineEuler
from physics_rk4 import PhysicsEngineRK4
from physics_leapfrog import PhysicsEngineLeap
from physics_skyfield import PhysicsEngineSkyfield
from render import RenderEngineVPython
from skyfield.api import load
import datetime
import numpy as np
import sys

class WorldLogOutput(WorldOutput):
    
    def start(self):
        self.print_state()
        
    def step(self):
        self.print_state()
        
    def print_state(self):
        for key, wo in self.world.world_objects:
            print('{time},{key},{wo}'.format(
                    time=self.world.time, key=key, wo=wo))
            
class CompareSkyfieldOutput(WorldOutput):
    
    planets = None
    timescale = None
    
    file_path = None
    file_object = None

    def __init__(self, world, file_path):
        super(CompareSkyfieldOutput, self).__init__(world)
        self.planets = load('de421.bsp')
        self.timescale = load.timescale()
        self.file_path = file_path
            
    def start(self):
        f = self.open_file()
        f.write('time')
        for key in self.world.world_objects.keys():
            f.write(' ' + key.replace(' ', '_'))
        f.write('\n')
        self.compare_timestamp()
        
    def step(self):
        self.compare_timestamp()
        
    def stop(self):
        self.close_file()
        
    def compare_timestamp(self):
        f = self.open_file()
        
        pydt = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc) + datetime.timedelta(seconds=self.world.time)
        t = self.timescale.utc(pydt)
        
        f.write(str(self.world.time).replace('.', ','))
        for key, wo in self.world.world_objects.items():
            planet = self.planets[key]
        
            pos = planet.at(t).ecliptic_xyz().km * 1000 # convert to m
            dist = np.linalg.norm(wo.position - pos)
            
            f.write(' ' + str(dist).replace('.', ','))
        f.write('\n')
        
    def open_file(self):
        # Reopen file if closed
        if self.file_object is None or self.file_object.closed:
            self.file_object = open(self.file_path, 'w+')
        return self.file_object
    
    def close_file(self):
        if self.file_object is not None:
            self.file_object.close()
            
            
input_type = sys.argv[1]
output_types = sys.argv[2].split(',')
end_time = float(sys.argv[3])
path = sys.argv[4]
world = World()

if input_type.startswith('file'):
    world.input = WorldFileInput(world, path)
elif input_type.startswith('euler'):
    #Init world
    o = WorldFileOutput(world, path)
    parts = input_type.split(':')
    dt = 1.0
    if len(parts) > 1:
        dt = float(parts[1])
    world.input = PhysicsEngineEuler(world, dt)
elif input_type.startswith('rk4'):
    #Init world
    o = WorldFileOutput(world, path)
    parts = input_type.split(':')
    dt = 1.0
    if len(parts) > 1:
        dt = float(parts[1])
    world.input = PhysicsEngineRK4(world, dt)
elif input_type.startswith('leapfrog'):
    #Init world
    o = WorldFileOutput(world, path)
    parts = input_type.split(':')
    dt = 1.0
    if len(parts) > 1:
        dt = float(parts[1])
    world.input = PhysicsEngineLeap(world, dt)
elif input_type.startswith('skyfield'):
    #Init world
    o = WorldFileOutput(world, path)
    parts = input_type.split(':')
    dt = 1.0
    if len(parts) > 1:
        dt = float(parts[1])
    world.input = PhysicsEngineSkyfield(world, dt)
        
for output_type in output_types:
    if output_type.startswith('vpython'):
        parts = output_type.split(':')
        time_scale = 1.0
        radius_scale = 1.0
        if len(parts) > 1:
            time_scale = float(parts[1])
        if len(parts) > 2:
            radius_scale = float(parts[2])
        world.outputs.append(RenderEngineVPython(world, time_scale, radius_scale))
    if output_type.startswith('file'):
        world.outputs.append(WorldFileOutput(world, path))
    if output_type.startswith('log'):
        world.outputs.append(WorldLogOutput(world))
    if output_type.startswith('compare_skyfield'):
        parts = output_type.split(':')
        file_path = 'compare_skyfield.csv'
        if len(parts) > 1:
            file_path = parts[1]
        world.outputs.append(CompareSkyfieldOutput(world, file_path))
    
world.start(end_time)